@@ -1,26 +1,59 @@
+from __future__ import annotations
+
 import warnings
+from abc import ABC, abstractmethod
 from copy import deepcopy
 from time import time
-from abc import ABC, abstractmethod
+from typing import Optional
+
 import numpy as np
-from matplotlib import pyplot as plt, ticker
+import numpy.typing as npt
+from matplotlib import pyplot as plt
+from matplotlib import ticker
 from scipy.optimize import fmin_slsqp
 from scipy.signal import find_peaks
 from tabulate import tabulate
 
-from autoeq.constants import DEFAULT_SHELF_FILTER_MIN_FC, DEFAULT_SHELF_FILTER_MAX_FC, DEFAULT_SHELF_FILTER_MIN_Q, \
-    DEFAULT_SHELF_FILTER_MAX_Q, DEFAULT_SHELF_FILTER_MIN_GAIN, DEFAULT_SHELF_FILTER_MAX_GAIN, \
-    DEFAULT_PEAKING_FILTER_MIN_FC, DEFAULT_PEAKING_FILTER_MAX_FC, DEFAULT_PEAKING_FILTER_MIN_Q, \
-    DEFAULT_PEAKING_FILTER_MAX_Q, DEFAULT_PEAKING_FILTER_MIN_GAIN, DEFAULT_PEAKING_FILTER_MAX_GAIN, \
-    DEFAULT_PEQ_OPTIMIZER_MIN_F, DEFAULT_PEQ_OPTIMIZER_MAX_F, DEFAULT_PEQ_OPTIMIZER_MAX_TIME, \
-    DEFAULT_PEQ_OPTIMIZER_TARGET_LOSS, DEFAULT_PEQ_OPTIMIZER_MIN_CHANGE_RATE, DEFAULT_PEQ_OPTIMIZER_MIN_STD
+from autoeq.constants import (
+    DEFAULT_PEAKING_FILTER_MAX_FC,
+    DEFAULT_PEAKING_FILTER_MAX_GAIN,
+    DEFAULT_PEAKING_FILTER_MAX_Q,
+    DEFAULT_PEAKING_FILTER_MIN_FC,
+    DEFAULT_PEAKING_FILTER_MIN_GAIN,
+    DEFAULT_PEAKING_FILTER_MIN_Q,
+    DEFAULT_PEQ_OPTIMIZER_MAX_F,
+    DEFAULT_PEQ_OPTIMIZER_MAX_TIME,
+    DEFAULT_PEQ_OPTIMIZER_MIN_CHANGE_RATE,
+    DEFAULT_PEQ_OPTIMIZER_MIN_F,
+    DEFAULT_PEQ_OPTIMIZER_MIN_STD,
+    DEFAULT_PEQ_OPTIMIZER_TARGET_LOSS,
+    DEFAULT_SHELF_FILTER_MAX_FC,
+    DEFAULT_SHELF_FILTER_MAX_GAIN,
+    DEFAULT_SHELF_FILTER_MAX_Q,
+    DEFAULT_SHELF_FILTER_MIN_FC,
+    DEFAULT_SHELF_FILTER_MIN_GAIN,
+    DEFAULT_SHELF_FILTER_MIN_Q,
+)
 
 
 class PEQFilter(ABC):
-    def __init__(self, f, fs,
-                 fc=None, optimize_fc=None, min_fc=None, max_fc=None,
-                 q=None, optimize_q=None, min_q=None, max_q=None,
-                 gain=None, optimize_gain=None, min_gain=None, max_gain=None):
+    def __init__(
+            self,
+            f: npt.NDArray[np.float64],
+            fs: int,
+            fc: Optional[float] = None,
+            optimize_fc: Optional[bool] = None,
+            min_fc: Optional[float] = None,
+            max_fc: Optional[float] = None,
+            q: Optional[float] = None,
+            optimize_q: Optional[bool] = None,
+            min_q: Optional[float] = None,
+            max_q: Optional[float] = None,
+            gain: Optional[float] = None,
+            optimize_gain: Optional[bool] = None,
+            min_gain: Optional[float] = None,
+            max_gain: Optional[float] = None
+    ) -> None:
         self.f = np.array(f)
         self._fs = fs
 
@@ -333,7 +366,12 @@ class HighShelf(ShelfFilter):
             # greatest
             min_ix = np.sum(self.f < max(40, self.min_fc))
             max_ix = np.sum(self.f < min(10000, self.max_fc))
-            ix = np.argmax([np.abs(np.mean(target[ix:])) for ix in range(min_ix, max_ix)])
+            # Phase 3 optimization: Vectorize mean calculation using cumsum
+            target_slice = target[min_ix:max_ix]
+            cumsum_rev = np.cumsum(target_slice[::-1])[::-1]
+            counts = np.arange(len(target_slice), 0, -1)
+            means = cumsum_rev / counts
+            ix = np.argmax(np.abs(means)) + min_ix
             self.fc = np.clip(self.f[ix], self.min_fc, self.max_fc)
             params.append(np.log10(self.fc))
         if self.optimize_q:
@@ -385,8 +423,12 @@ class LowShelf(ShelfFilter):
             # greatest
             min_ix = np.sum(self.f < max(40, self.min_fc))
             max_ix = np.sum(self.f < min(10000, self.max_fc))
-            ix = np.argmax([np.abs(np.mean(target[:ix + 1])) for ix in range(min_ix, max_ix)])
-            ix += min_ix
+            # Phase 3 optimization: Vectorize mean calculation using cumsum
+            target_slice = target[min_ix:max_ix]
+            cumsum_fwd = np.cumsum(target_slice)
+            counts = np.arange(1, len(target_slice) + 1)
+            means = cumsum_fwd / counts
+            ix = np.argmax(np.abs(means)) + min_ix
             self.fc = np.clip(self.f[ix], self.min_fc, self.max_fc)
             params.append(np.log10(self.fc))
         if self.optimize_q:
@@ -716,7 +758,7 @@ class PEQ:
                 iter=1000,  # 최대 반복 횟수 명시
                 acc=1e-6,   # 수렴 정확도 명시
                 iprint=0)
-        except OptimizationFinished as err:
+        except OptimizationFinished:
             # Restore best params
             self._parse_optimizer_params(self.history.params[np.argmin(self.history.loss)])
 
